# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import time


class Timer(object):

    def __init__(self):
        self.start = time.perf_counter()
        self.prev = self.start

    def tic(self):
        prev = self.prev
        self.prev = time.perf_counter()
        return self.prev - prev, self.prev - self.start

    def tic_print(self):
        split, total = self.tic()
        print(f"- Elapsed: {split}s [Total: {total}s]")



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse

from azulero import assemble, retrieve, process, crop


def run():

    parser = argparse.ArgumentParser(
        prog="azul",
        description="Bring colors to Euclid tiles!",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--workspace", type=str, default=".", metavar="PATH", help="Parent workspace"
    )

    subparsers = parser.add_subparsers(title="Commands")
    retrieve.add_parser(subparsers)
    crop.add_parser(subparsers)
    process.add_parser(subparsers)
    assemble.add_parser(subparsers)

    args = parser.parse_args()
    args.func(args)



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
import requests

from azulero import io
from azulero.timing import Timer


class DSS(object):

    def query_datafiles(self, tile, dsr):

        query = {
            "project": "EUCLID",
            "class_name": "DpdMerBksMosaic",
            "Data.TileIndex": tile,
            "Header.DataSetRelease": dsr,
            "fields": "Data.DataStorage.DataContainer.FileName:Data.Filter.Name",
        }
        lines = (
            requests.get("https://eas-dps-rest-ops.esac.esa.int/REST", params=query)
            .text.replace('"', "")
            .split()
        )
        datafiles = {}
        for l in lines:
            if "VIS" in l or "NIR" in l:
                file_name, filter_name = l.split(",")
                datafiles[file_name] = filter_name
        return datafiles

    def download_datafile(self, name, path):

        r = requests.get(f"https://euclidsoc.esac.esa.int/{name}")
        with open(path, "wb") as f:
            f.write(r.content)


class SAS(object):

    def query_datafiles(self, tile, dsr):
        adql = (
            f"SELECT TOP 50 file_name, filter_name FROM sedm.mosaic_product"
            f" WHERE (release_name='{dsr}')"
            f" AND (category='SCIENCE')"
            f" AND (tile_index={tile})"
            f" AND (instrument_name IN ('VIS', 'NISP'))"
        )
        query = {
            "REQUEST": "doQuery",
            "LANG": "ADQL",
            "FORMAT": "csv",
            "QUERY": adql.replace(" ", "+"),
        }
        url = "https://eas.esac.esa.int/tap-server/tap/sync?" + "&".join(
            f"{p}={query[p]}" for p in query
        )
        r = requests.get(url)  # Cannot use params as adql characters would be escaped

        lines = r.text.split()
        datafiles = {}
        for l in lines[1:]:
            file_name, filter_name = l.split(",")
            datafiles[file_name] = filter_name
        return datafiles

    def download_datafile(self, name, path):

        query = {"file_name": name, "release": "sedm", "RETRIEVAL_TYPE": "FILE"}
        r = requests.get(f"https://eas.esac.esa.int/sas-dd/data", query)
        with open(path, "wb") as f:
            f.write(r.content)


providers = {"dss": DSS, "sas": SAS}


def enumeration(values, coordination=", "):
    l = [str(v) for v in values]
    if len(l) == 1:
        return l[0]
    return ", ".join(list(l)[:-1]) + coordination + list(l)[-1]


def choice(values):
    return enumeration(values, " or ")


def add_parser(subparsers):

    parser = subparsers.add_parser(
        "retrieve",
        help="Retrieve MER datafiles.",
        description="TODO",  # FIXME
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "tiles",
        type=str,
        nargs="+",
        metavar="INDICES",
        help="Space-separated list of tile indices.",
    )
    parser.add_argument(
        "--dsr",
        type=str,
        default="DR1_R2,DR1_R1,Q1_R1",
        help="Comma-separated list of data set releases.",
    )
    parser.add_argument(
        "--from",
        type=str,
        default="dss",
        metavar="PROVIDER",
        help=f"Data provider: {choice(providers.keys())}.",
    )

    parser.set_defaults(func=run)


def query_datafiles(retriever, tile, dsr):

    print(f"Query datafiles for tile {tile} and dataset release {dsr}:")

    datafiles = retriever.query_datafiles(tile, dsr)
    if len(datafiles) == 0:
        print("- None found.")

    for f in datafiles:
        print(f"- [{datafiles[f]}] {f}")
    return datafiles


def download_datafiles(retriever, datafiles, workdir):

    print(f"Download and extract datafiles to: {workdir}")

    for name in datafiles:  # TODO parallelize?
        path = workdir / name.removesuffix(".gz")
        if path.is_file():
            print(f"WARNING: File exists; skip: {path.name}")
            continue
        retriever.download_datafile(name, path)
        print(f"- {path}")


def run(args):

    print()

    timer = Timer()
    provider = providers[vars(args)["from"]]()
    for tile in args.tiles:  # TODO parallelize?
        workdir = io.make_workdir(args.workspace, tile)
        for dsr in args.dsr.split(","):
            datafiles = query_datafiles(provider, tile, dsr)
            if len(datafiles) > 0:
                break
        timer.tic_print()
        if len(datafiles) < 4:
            print(f"ERROR: Only {len(datafiles)} files found; Skipping this tile.")
            continue
        if len(datafiles) > 4:
            print(f"WARNING: More than 4 files found: {len(datafiles)}.")
        download_datafiles(provider, datafiles, workdir)
        timer.tic_print()
        print(f"\nYou may now run:")
        print(f"\nazul --workspace {args.workspace} crop {tile}\n")
        print(f"or:")
        print(f"\nazul --workspace {args.workspace} process {tile}\n")



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
import numpy as np
from pathlib import Path
import yaml

from azulero import color, io, mask
from azulero.timing import Timer


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "assemble",
        help="Assemble tile patches for testing purposes.",
        description=(
            "Assemble tile patches in a grid, "
            "in order to process a varied collection of objects and tune parameters."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "tiles",
        type=str,
        metavar="SPECS",
        help="Path to the YAML configuration file (list of tile specs), relative to the workspace",
    )
    # parser.add_argument( # FIXME
    #     "--cols", "-c", metaval="COUNT", help="Maximum number of columns"
    # )
    parser.add_argument(
        "--output-dir",
        "-o",
        type=str,
        default="ASSEMBLAGE",
        metavar="PATH",
        help="Output assemblage directory.",
    )

    parser.set_defaults(func=run)


def height(slicing):
    vertical = slicing[0]
    return vertical.stop - vertical.start


def run(args):

    print()

    timer = Timer()
    workspace = Path(args.workspace).expanduser()

    print("Read patches")
    patches = []
    with open(workspace / args.tiles, "r") as f:
        specs = [io.parse_tile(tile) for tile in yaml.safe_load(f)]
    common_height = height(specs[0][1])
    print(f"- Height: {common_height}")
    assert all(
        height(slicing) == common_height for _, slicing in specs
    ), "All patches must have the same height."
    for tile, slicing in specs:
        workdir = workspace / tile
        patch = io.read_iyjh(workdir, slicing)
        print(f"- {tile}: {patch.shape[1]} x {patch.shape[2]}")
        patches.append(patch)
        timer.tic_print()

    print("Assemble")
    assemblage = np.concatenate(patches, axis=2)
    print(f"- Shape: {assemblage.shape[1]} x {assemblage.shape[2]}")
    timer.tic_print()

    print("Write channels")
    workdir = io.make_workdir(workspace, args.output_dir)
    for name, channel in zip(("VIS", "NIR-Y", "NIR-J", "NIR-H"), assemblage):
        path = workdir / f"EUC_{name}_ASSEMBLAGE.fits"
        print(f"- [{name}] {path}")
        io.write_fits(channel, path)
    timer.tic_print()



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
from astropy.io import fits
import csv


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "tiling",
        type=str,
        help="Tile catalog with list of noteworthy objects",
    )
    parser.add_argument("output", type=str, help="Output objects catalog")
    return parser.parse_args()


def process(tiling):

    print(f"Reading tile catalog: {tiling}")
    hdul = fits.open(tiling)
    data = hdul[1].data
    tiles = data["tileId"]
    objects = data["comment"]
    print(f"- Tiles: {len(tiles)}")
    res = {int(t): "".join(o).split(",") for t, o in zip(tiles, objects) if len(o) > 1}
    print(f"- Objects: {len(res)}")
    return res


def write(catalog, output):
    print(f"Writing object catalog: {output}")
    with open(output, "w") as f:
        w = csv.writer(f, delimiter="\t")
        w.writerow(["Tile", "Count", "Objects"])
        for t in catalog:
            objects = catalog[t]
            w.writerow([str(t), len(objects), ", ".join(objects)])


if __name__ == "__main__":

    args = parse_args()
    catalog = process(args.tiling)
    write(catalog, args.output)
